from pprint import pprint

from .response import Response
from .request import Request
from .transport import Urllib3Transport
from .error import NetworkError


class Session(object):
    def __init__(self):
        self.transport = Urllib3Transport()

    def request(self, url=None, **kwargs):
        req = Request(url=url, **kwargs)
        res = Response()
        self.transport.prepare_request(req, res)
        self.transport.request(req, res)
        self.transport.prepare_response(req, res, None)
        return res


def request(*args, **kwargs):
    sess = Session()
    return sess.request(*args, **kwargs)



from pprint import pprint
import logging
import time
from contextlib import contextmanager
import traceback
import sys
from urllib.parse import urlencode
import ssl
import re
from functools import partial

from urllib3.filepost import encode_multipart_formdata
from urllib3.util.retry import Retry
from urllib3.util.timeout import Timeout
from urllib3 import exceptions, ProxyManager, make_headers
from urllib3.contrib import pyopenssl
from urllib3.contrib.socks import SOCKSProxyManager
import urllib3
import OpenSSL.SSL
import certifi

from . import error
from .urllib3_custom import CustomPoolManager

urllib3.disable_warnings(exceptions.InsecureRequestWarning)
pyopenssl.inject_into_urllib3()

# Fix processing urls which end with "?"
urllib3_ver = tuple(int(x) for x in urllib3.__version__.split('.'))
if urllib3_ver <= (1, 25, 6): 
    import urllib3.util.url
    urllib3.util.url.TARGET_RE = re.compile(r"^(/[^?]*)(?:\?([^#]*))?(?:#(.*))?$")


class Urllib3Transport(object):
    __slots__ = (
        'urllib3_response',
        'op_started',
        'pools',
        'prepare_response_hook',
    )

    def __init__(
            self,
            prepare_response_hook=None,
            num_pools=10,
            pool_size=1,
        ):
        self.prepare_response_hook = prepare_response_hook
        self.urllib3_response = None
        self.pools = {
            (None, None, True): CustomPoolManager(
                cert_reqs='CERT_REQUIRED',
                ca_certs=certifi.where(),
                num_pools=num_pools,
                maxsize=pool_size,
            ),
            (None, None, False): CustomPoolManager(
                cert_reqs='CERT_NONE',
                num_pools=num_pools,
                maxsize=pool_size,
            ),
        }



    def prepare_request(self, req, res):
        pass

    @contextmanager
    def handle_network_error(self, req):
        try:
            yield
        except exceptions.ReadTimeoutError as ex:
            raise error.OperationTimeoutError(str(ex), ex)
        except exceptions.ConnectTimeoutError as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.ProtocolError as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.SSLError as ex:
            raise error.ConnectError(str(ex), ex)
        except OpenSSL.SSL.Error as ex:
            raise error.ConnectError(str(ex), ex)
        except ssl.SSLError as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.LocationParseError as ex:
            raise error.MalformedResponseError(str(ex), ex)
        except exceptions.LocationValueError as ex:
            raise error.InvalidUrlError(str(ex), ex)
        except exceptions.DecodeError as ex:
            raise error.MalformedResponseError(str(ex), ex)
        except exceptions.InvalidHeader as ex:
            raise error.MalformedResponseError(str(ex), ex)
        except exceptions.ProxyError as ex:
            raise error.ProxyError(str(ex), ex)
        except exceptions.MaxRetryError as ex:
            # Might be raised by multiple reasons
            # So we just raise original error and process it
            # with `self.handle_network_error()` once again
            with self.handle_network_error(req):
                raise ex.reason
        except exceptions.ResponseError as ex:
            if 'too many redirects' in str(ex):
                raise error.TooManyRedirectsError(str(ex), ex)
            else:
                raise
        except AttributeError:
            # See https://github.com/urllib3/urllib3/issues/1556
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            found = False
            for frame in frames:
                if (
                        "if host.startswith('[')" in frame.line
                        and 'connectionpool.py' in frame.filename
                    ):
                    found = True
                    break
            if found:
                raise error.MalformedResponseError('Invalid redirect header')
            else:
                raise
        except UnicodeError as ex:
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            #    for res in _socket.getaddrinfo(host, port, family, type, proto, flags):
            #UnicodeError: encoding with 'idna' codec failed (UnicodeError: label empty or too long)
            if "encoding with 'idna' codec failed" in str(ex):
                raise error.InvalidUrlError(
                    'Fail to process redirect from %s' % req.config.get('url')
                )
            else:
                raise
        except UnicodeEncodeError as ex:
            # UnicodeEncodeError is subclass of ValueError
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            #    self._output(request.encode('ascii'))
            #UnicodeEncodeError: 'ascii' codec can't encode characters
            #in position 15-24: ordinal not in range(128)
            if 'self._output(request.encode' in frames[-1].line:
                raise error.InvalidUrlError(
                    'Non ASCII URL: %s' % req.config.get('url')
                )
            else:
                raise
        except ValueError as ex:
            if 'Invalid IPv6 URL' in str(ex):
                raise error.MalformedResponseError('Invalid redirect header')
            else:
                raise
        except KeyError as ex:
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            if 'self.key_fn_by_scheme[scheme]' in frames[-1].line:
                raise error.InvalidUrlError(
                    'Unknown URL scheme: %s' % req.config.get('url')
                )
            else:
                raise



    def get_pool(self, req, use_cache=True):
        if req['proxy']:
            if req['proxy_type'] == 'socks5' and req['proxy_auth']:
                proxy_url = '%s://%s@%s' % (
                    req['proxy_type'], req['proxy_auth'], req['proxy']
                )
            else:
                proxy_url = '%s://%s' % (req['proxy_type'], req['proxy'])
            pool_key = (req['proxy_type'], req['proxy'], bool(req['verify']))
            if not use_cache or pool_key not in self.pools:
                if req['proxy_type'] == 'socks5':
                    if req['verify']:
                        pool = SOCKSProxyManager(
                            proxy_url,
                            cert_reqs='CERT_REQUIRED',
                            ca_certs=certifi.where(),
                        )
                    else:
                        pool = SOCKSProxyManager(proxy_url, **opts)
                elif req['proxy_type'] == 'http':
                    if req['proxy_auth']:
                        proxy_headers = make_headers(proxy_basic_auth=req['proxy_auth'])
                    else:
                        proxy_headers = None
                    if req['verify']:
                        pool = ProxyManager(
                            proxy_url,
                            proxy_headers=proxy_headers,
                            cert_reqs='CERT_REQUIRED',
                            ca_certs=certifi.where(),
                        )
                    else:
                        pool = ProxyManager(
                            proxy_url,
                            proxy_headers=proxy_headers,
                        )
                else:
                    raise error.IowebConfigError(
                        'Invalid value of request option `proxy_type`: %s'
                        % req['proxy_type']
                    )
                if use_cache:
                    self.pools[pool_key] = pool
            else:
                pool = self.pools[pool_key]
        else:
            pool = self.pools[(None, None, bool(req['verify']))]
        return pool

    def request(self, req, res):
        options = {}
        headers = req['headers'] or {}

        req_url = req['url']
        if req_url.endswith('?'):
            req_url = req['url'][:-1]

        #req_url = req['url']
        #if not '://' in req_url:
        #    req_url = 'http://%s' % req_url
        #url_parts = urlsplit(req_url)
        #if url_parts.scheme not in ('http', 'https'):
        #    emsg = 'Unknown scheme [%s]' % url_parts.scheme
        #    # second argument saved as `err.transport_error`
        #    # and used then to generate short code of error
        #    # for statistics
        #    raise error.InvalidUrlError(
        #        emsg, error.InvalidUrlError(emsg),
        #    )

        pool = self.get_pool(
            req,
            use_cache=(not req['close_connection'])
        )

        self.op_started = time.time()
        if req['resolve']:
            if req['proxy']:
                raise error.IowebConfigError(
                    'Request option `resolve` could not be used along option `proxy`'
                )
            for host, ip in req['resolve'].items():
                pool.resolving_cache[host] = ip


        if req['content_encoding']:
            if not any(x.lower() == 'accept-encoding' for x in headers):
                headers['Accept-Encoding'] = req['content_encoding']

        if req['data']:
            if isinstance(req['data'], dict):
                if req['multipart']:
                    body, ctype = encode_multipart_formata(req['data'])
                else:
                    body = urlencode(req['data'])
                    ctype = 'application/x-www-form-urlencoded'
                options['body'] = body
                headers['Content-Type'] = ctype
            elif isinstance(req['data'], bytes):
                options['body'] = req['data']
            elif isinstance(req['data'], str):
                options['body'] = req['data'].encode('utf-8')
            else:
                raise error.IowebConfigError(
                    'Invalid type of request data option: %s'
                    % type(req['data'])
                )
            headers['Content-Length'] = len(options['body'])

        with self.handle_network_error(req):
            retry_opts = {
                # total - set to None to remove this constraint
                # and fall back on other counts. 
                'connect': False,
                'read': False,
            }
            if req['follow_redirect']:
                retry_opts.update({
                    'total': req['max_redirects'],
                    'redirect': req['max_redirects'],
                    'raise_on_redirect': True,
                })
            else:
                retry_opts.update({
                    'total': False,
                    'redirect': False,
                    'raise_on_redirect': False,
                })
            self.urllib3_response = pool.urlopen(
                req.method(),
                req_url,
                headers=headers,
                retries=Retry(**retry_opts),
                timeout=Timeout(
                    connect=req['connect_timeout'],
                    read=req['timeout'],
                ),
                preload_content=False,
                decode_content=req['decode_content'],
                **options
            )

    def read_with_timeout(self, req, res):
        read_limit = req['content_read_limit']
        if read_limit:
            read_limit = int(read_limit)
        chunk_size = (2**16) # = 65536
        bytes_read = 0
        chunks = []
        try:
            while True:
                chunk = self.urllib3_response.read(chunk_size)
                if chunk:
                    if read_limit:
                        chunk_limit = min(len(chunk), read_limit - bytes_read)
                    else:
                        chunk_limit = len(chunk)
                    chunks.append(chunk[:chunk_limit])
                    bytes_read += chunk_limit
                    if read_limit and bytes_read >= read_limit:
                        break
                else:
                    break
                if time.time() - self.op_started > req['timeout']:
                    raise error.OperationTimeoutError(
                        'Timed out while reading response',
                    )
        finally:
            res._bytes_body = b''.join(chunks)

    def prepare_response(self, req, res, err, raise_network_error=True):
        try:
            if err:
                res.error = err
            else:
                try:
                    res.headers = self.urllib3_response.headers
                    res.status = self.urllib3_response.status
                    if len(self.urllib3_response.retries.history):
                        res.url = (
                            # redirect_location might be None
                            # if network errors happened during
                            # processing redirect
                            self.urllib3_response.retries.history[-1].redirect_location
                            or
                            self.urllib3_response.retries.history[-1].url
                        )
                    else:
                        res.url = req['url']
                        
                    if (
                            req['extract_cert']
                            and self.urllib3_response._connection.sock
                        ):
                        res.cert = (
                            self.urllib3_response._connection.sock.connection
                            .get_peer_cert_chain()
                        )

                    if self.prepare_response_hook:
                        self.prepare_response_hook(
                            self, req, res, self.urllib3_response
                        )

                    with self.handle_network_error(req):
                        try:
                            self.read_with_timeout(req, res)
                        except MemoryError as ex:
                            raise error.MalformedResponseError(
                                'Could not decompress response content',
                                ex
                            )

                except (error.NetworkError, error.MalformedResponseError) as ex:
                    if raise_network_error:
                        raise
                    else:
                        res.error = err
        finally:
            if self.urllib3_response:
                #if req['close_connection']:
                #    self.urllib3_response._connection.close()
                self.urllib3_response.release_conn()



from pprint import pprint
import time
#from collections import defaultdict

from pymongo import UpdateOne
from pymongo.errors import BulkWriteError


def bulk_write(db, item_type, ops, stat=None, retries=3):
    """
    Tries to apply `ops` Update operations to `item_type`
    collection with `bulk_write` method.
    Gives up if `retries` retries failed.

    Args:
        db - pymongo collection object
        item_type - name of collection
        ops - list of operations like UpdateOne
        stat - instance of `ioweb.stat:Stat`
        retries - number of retries
    """
    if stat:
        stat.inc('bulk-write-%s' % item_type)
    for retry in range(retries):
        try:
            res = db[item_type].bulk_write(ops, ordered=False)
        except BulkWriteError as ex:
            if retry == (retries - 1):
                raise
            else:
                if stat:
                    stat.inc('bulk-write-%s-retry' % item_type)
        else:
            if stat:
                stat.inc(
                    'bulk-write-%s-upsert' % item_type,
                    res.bulk_api_result['nUpserted']
                )
                stat.inc(
                    'bulk-write-%s-change' % item_type,
                    res.bulk_api_result['nModified']
                )
            break
    return res


class BulkWriter(object):
    def __init__(self, db, item_type, bulk_size=100, stat=None, retries=3):
        self.db = db
        self.item_type = item_type
        self.stat = stat
        self.retries = retries
        self.bulk_size = bulk_size
        self.ops = []

    def _write_ops(self):
        bulk_write(self.db, self.item_type, self.ops, self.stat)
        self.ops = []

    def update_one(self, *args, **kwargs):
        self.ops.append(UpdateOne(*args, **kwargs))
        if len(self.ops) >= self.bulk_size:
            self._write_ops()

    def flush(self):
        if len(self.ops):
            self._write_ops()


def iterate_collection(
        db, item_type, query, sort_field, iter_chunk=1000,
        fields=None, infinite=False, limit=None
    ):
    """
    Iterate over `db[item_type]` collection items matching `query`
    sorted by `sort_field`.

    Intenally, it fetches chunk of `iter_chunk` items at once and
    iterates over it. Then fetch next chunk.
    """
    recent_id = None
    count = 0
    while True:
        if recent_id:
            query[sort_field] = {'$gt': recent_id}
        items = list(db[item_type].find(
            query, fields, sort=[(sort_field, 1)], limit=iter_chunk
        ))
        if not items:
            if infinite:
                sleep_time = 5
                logging.debug(
                    'No items to process. Sleeping %d seconds'
                    % sleep_time
                )
                time.sleep(sleep_time)
                recent_id = None
            else:
                return
        else:
            for item in items:
                yield item
                recent_id = item[sort_field]
                count += 1
                if limit and count >= limit:
                    return



import time
import sys
import logging
from threading import Event, Thread, Lock
from queue import Queue, PriorityQueue, Empty, Full
from urllib.parse import urlsplit 
import re
#import json
from bson import json_util as json
from urllib.request import urlopen
from traceback import format_exception
from collections import defaultdict
import gc
from copy import deepcopy
from datetime import datetime

from .util import Pause, debug
#from .loop import MultiCurlLoop
from .network_service import NetworkService
from .stat import Stat
from .request import Request, CallbackRequest, BaseRequest
from .error import (
    get_error_tag, collect_error_context, DataNotValid,
    IowebConfigError
)
from .error_logger import ErrorLogger
from .proxylist import ProxyList


class Crawler(object):
    _taskgen_sleep_time = 0.01
    dataop_threshold_default = {
        'number': 500,
        'size': None,
        #'time': None,
    }
    dataop_threshold = {}

    def task_generator(self):
        if False:
            yield None

    def __init__(self,
            network_threads=3,
            result_workers=4,
            num_task_generators=1,
            retry_limit=3,
            extra_data=None,
            stop_on_handler_error=False,
            debug=False,
            stat_logging=True,
            stat_logging_format='text',
            master_taskq=None
        ):
        if extra_data is None:
            self.extra_data = {}
        else:
            self.extra_data = deepcopy(extra_data)
        self.taskq = PriorityQueue()
        self.taskq_size_limit = max(100, network_threads * 2)
        self.resultq = Queue()
        self.shutdown_event = Event()
        self.fatal_error_happened = Event()
        self.network_pause = Pause()
        self.retry_limit = retry_limit
        self.stat = Stat(
            speed_keys='crawler:request-processed',
            logging_enabled=stat_logging,
            logging_format=stat_logging_format,
        )
        self.fatalq = Queue()
        self.network = NetworkService(
            self.taskq, self.resultq,
            fatalq=self.fatalq,
            threads=network_threads,
            shutdown_event=self.shutdown_event,
            pause=self.network_pause,
            setup_request_hook=self.setup_request_hook,
            setup_request_proxy_hook=self.setup_request_proxy_hook,
            prepare_response_hook=self.prepare_response_hook,
            stat=self.stat,
        )
        self.result_workers = result_workers
        self._run_started = None

        self.dataopq = {}
        self.dataop_lock = defaultdict(Lock)
        self.dataop_counters = defaultdict(lambda: {
            'number': 0,
            'size': 0,
        })
        self.error_logger = ErrorLogger()
        self.stop_on_handler_error = stop_on_handler_error
        self.proxylist = None
        self.debug = debug
        self.num_task_generators = num_task_generators
        self.master_taskq_cfg = master_taskq

        self.init_hook()

    @classmethod
    def update_arg_parser(cls, parser):
        for key, config in cls.extra_cli_args().items():
            parser.add_argument('--extra-%s' % key, **config)

    @classmethod
    def extra_cli_args(cls):
        return {}

    def load_proxylist(self, list_type, list_location, **kwargs):
        self.proxylist = ProxyList.create_from_source(
            list_type, list_location, **kwargs
        )

    def is_dataopq_dump_time(self, name):
        to_dump = False
        try:
            number_th = self.dataop_threshold[name]['number']
        except KeyError:
            number_th = self.dataop_threshold_default['number']
        if number_th:
            if self.dataop_counters[name]['number'] >= number_th:
                    to_dump = True
        if not to_dump:
            try:
                size_th = self.dataop_threshold[name]['size']
            except KeyError:
                size_th = self.dataop_threshold_default['size']
            if size_th:
                if self.dataop_counters[name]['size'] >= size_th:
                    to_dump = True
        return to_dump

    def enq_dataop(self, name, op, size=None, force_dump=False):
        if op:
            self.enq_dataop_hook(name, op)
        self.dataop_lock[name].acquire()
        released = False
        try:
            self.dataopq.setdefault(name, [])
            if op:
                self.dataopq[name].append(op)
                self.dataop_counters[name]['number'] += 1
                if size:
                    self.dataop_counters[name]['size'] += size

            if force_dump or self.is_dataopq_dump_time(name):
                #logging.debug('Dumping data ops')
                self.stat.inc('dataop-dump-%s' % name)
                ops = self.dataopq[name]
                if ops:
                    self.dataopq[name] = []
                    self.dataop_counters[name]['number'] = 0
                    self.dataop_counters[name]['size'] = 0
                    # release as soon as possible
                    self.dataop_lock[name].release()
                    released = True
                    func = getattr(self, 'dataop_handler_%s' % name)
                    func(ops)
                else:
                    self.dataop_lock[name].release()
                    released = True
        finally:
            if not released:
                self.dataop_lock[name].release()

    def enq_dataop_hook(self, name, op):
        pass

    def init_hook(self):
        pass

    def prepare_response_hook(self, transport, req, res, urllib3_response):
        pass

    def setup_request_proxy_hook(self, transport, req):
        if self.proxylist:
            proxy = self.proxylist.random_server()
            req.setup(
                proxy=proxy.address(),
                proxy_auth=proxy.auth(),
                proxy_type=proxy.proxy_type,
            )
            # If requst is configured to not close connection (default)
            # and proxylist requires to close connection
            # then update request config
            if (
                    not req['close_connection']
                    and self.proxylist.close_connection
                ):
                req.setup(close_connection=True)

    def setup_request_hook(self, transport, req):
        pass

    def submit_task(self, req):
        self.submit_task_hook(req)
        self.taskq.put((req.priority, req))

    def submit_task_hook(self, req):
        pass

    def task_generator_master(self):
        cfg = self.master_taskq_cfg
        if not cfg:
            raise IowebConfigError('Master task queue is not configured')
        backend_id = cfg.pop('backend', None)
        if backend_id not in ('redis',):
            raise IowebConfigError(
                'Invalid master task queue backend: %s' % backend_id
            )
        # Redis specific things
        # TODO: move into backend class
        taskq_id = cfg.pop('taskq_id', None)
        if not taskq_id:
            raise IowebConfigError(
                'Invalid queue ID: %s' % taskq_id
            )
        from redis import Redis
        rdb = Redis(decode_responses=True, **cfg)
        DONE = False
        while True:
            task = rdb.lpop(taskq_id)
            if task is None:
                time.sleep(0.01)
            else:
                task = json.loads(task)
                if '$op' in task:
                    if task['$op'] == 'shutdown':
                        return
                    else:
                        raise IowebConfigError(
                            'Got invalid $op from master taskq queue: %s'
                            % task['$op']
                        )
                else:
                    req = Request.from_data(task)
                    yield req

    def thread_task_generator(self):
        try:
            try:
                if self.master_taskq_cfg:
                    tgen_iter = self.task_generator_master()
                else:
                    tgen_iter = iter(self.task_generator())
            except TypeError:
                return
            else:
                for item in tgen_iter:
                    while item:
                        if self.shutdown_event.is_set():
                            return
                        if self.taskq.qsize() >= self.taskq_size_limit:
                            time.sleep(self._taskgen_sleep_time)
                        else:
                            self.submit_task(item)
                            item = None
        except (Exception, KeyboardInterrupt) as ex:
            self.fatalq.put((sys.exc_info(), None))

    def is_result_ok(self, req, res):
        if res.error:
            return False
        elif isinstance(req, CallbackRequest):
            return True
        elif (
                0 < res.status < 400
                or res.status == 404
                or (
                    req.config['extra_valid_status']
                    and res.status in req.config['extra_valid_status']
                )
            ):
            return True
        else:
            return False

    def thread_result_processor(self, pause):
        try:
            error_ctx = None
            while not self.shutdown_event.is_set():
                if pause.pause_event.is_set():
                    pause.process_pause()
                try:
                    result = self.resultq.get(True, 0.1)
                except Empty:
                    pass
                else:
                    self.stat.inc('crawler:request-processed')
                    try:
                        if (
                                result['request']['raw']
                                or self.is_result_ok(
                                    result['request'],
                                    result['response'],
                                )
                            ):
                            self.process_ok_result(result)
                        else:
                            self.process_fail_result(result)
                    except Exception as ex:
                        error_ctx = collect_error_context(result['request'])
                        raise
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), error_ctx))

    def thread_fatalq_processor(self):
        try:
            while not self.shutdown_event.is_set():
                try:
                    exc_info, ctx = self.fatalq.get(True, 0.1)
                except Empty:
                    pass
                else:
                    self.shutdown_event.set()
                    self.fatal_error_happened.set()
                    logging.error('Fatal exception')
                    if ctx:
                        logging.error('Exception context:')
                        for key, val in sorted(ctx.items(), key=lambda x: x[0]):
                            logging.error(' * %s: %s' % (key, val))
                    else:
                        logging.error('Exception context: N/A')
                    logging.error(''.join(format_exception(*exc_info)))
                    if not isinstance(exc_info[1], KeyboardInterrupt):
                        self.log_error(exc_info, ctx)
        except Exception as ex:
            self.shutdown_event.set()
            raise

    def log_error(self, exc_info, ctx=None):
        ctx = ctx or {}
        ctx.update({
            'crawler_id': self.__class__.__name__,
            'date': datetime.utcnow().isoformat(),
        })
        self.error_logger.log_error(exc_info, ctx)

    def thread_manager(self, task_gens, pauses):
        try:
            [x.join() for x in task_gens]

            def system_is_busy():
                return (
                    self.taskq.qsize()
                    or self.resultq.qsize()
                    or len(self.network.active_handlers)
                )

            while not self.shutdown_event.is_set():
                while system_is_busy():
                    if self.shutdown_event.is_set():
                        return
                    time.sleep(0.1)

                for pause in pauses:
                    pause.pause_event.set()
                ok = True
                for pause in pauses:
                    if not pause.is_paused.wait(0.1):
                        ok = False
                        break
                if not ok:
                    for pause in pauses:
                        pause.pause_event.clear()
                        pause.resume_event.set()
                else:
                    if not system_is_busy():
                        for pause in pauses:
                            pause.pause_event.clear()
                            pause.resume_event.set()
                        self.shutdown_event.set()
                        return
                    else:
                        for pause in pauses:
                            pause.pause_event.clear()
                            pause.resume_event.set()
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), None))

    def process_ok_result(self, result):
        self.stat.inc('crawler:request-ok')
        name = result['request'].config['name']
        handler = getattr(self, 'handler_%s' % name)
        try:
            handler_result = handler(
                result['request'],
                result['response'],
            )
            try:
                iter(handler_result)
            except TypeError:
                pass
            else:
                for item in handler_result:
                    if isinstance(item, BaseRequest):
                        self.submit_task(item)
                    else:
                        raise Exception(
                            'Handler yielded non request task: %s' % item
                        )
        except Exception as ex:
            self.stat.inc('result-handler-error:%s' % get_error_tag(ex))
            if isinstance(ex, DataNotValid):
                self.process_fail_result(result)
            else:
                if self.stop_on_handler_error:
                    raise
                else:
                    logging.exception('Error in result handler')
                    ctx = collect_error_context(result['request'])
                    self.log_error(sys.exc_info(), ctx)

    def process_fail_result(self, result):
        self.stat.inc('crawler:request-fail')
        req = result['request']

        if result['response'].error:
            self.stat.inc('network-error:%s' % get_error_tag(
                result['response'].error
            ))
        if result['response'].status:
            self.stat.inc('http:status-%s' % result['response'].status)

        if req.retry_count + 1 < self.retry_limit:
            self.stat.inc('crawler:request-retry')
            req.retry_count += 1
            req.priority = req.priority - 1
            self.submit_task(req)
        else:
            self.stat.inc('crawler:request-rejected')
            name = result['request'].config['name']
            handler = getattr(self, 'rejected_handler_%s' % name, None)
            if handler is None:
                handler = self.default_rejected_handler
            try:
                handler(result['request'], result['response'])
            except Exception as ex:
                self.stat.inc('rejected-handler-error:%s' % get_error_tag(ex))
                if self.stop_on_handler_error:
                    raise
                else:
                    logging.exception('Error in rejected result handler')
                    ctx = collect_error_context(req)
                    self.log_error(sys.exc_info(), ctx)

    def default_rejected_handler(self, req, res):
        pass

    def run_hook(self):
        pass

    def thread_debug(self):
        try:
            while not self.shutdown_event.is_set():
                stat = []
                now = time.time()
                for hdl in self.network.active_handlers:
                    stat.append((hdl, self.network.registry[hdl]['start']))
                with open('var/crawler.debug', 'w') as out:
                    for hdl, start in list(sorted(stat, key=lambda x: (x[1] or now), reverse=True)):
                        req = self.network.registry[hdl]['request']
                        out.write('%.2f - [#%s] - %s\n' % (
                            (now - start) if start else 0,
                            req.retry_count if req else 'NA',
                            urlsplit(req['url']).netloc if req else 'NA',
                        ))
                    out.write('Active handlers: %d\n' % len(self.network.active_handlers))
                    out.write('Idle handlers: %d\n' % len(self.network.idle_handlers))
                    out.write('Taskq size: %d\n' % self.taskq.qsize())
                    out.write('Resultq size: %d\n' % self.resultq.qsize())

                total = 0
                count = 0
                for hdl, start in stat:
                    if start:
                        total += (now - start)
                        count += 1
                logging.debug('Median handler time: %.2f' % ((total / count) if count else 0))

                self.shutdown_event.wait(3)
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), None))


    def thread_network(self):
        try:
            self.network.run()
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), None))

    def shutdown(self):
        for name in self.dataopq.keys():
            self.enq_dataop(name, None, force_dump=True)
        self.shutdown_hook()

    def shutdown_hook(self):
        pass

    def run(self):
        try:
            self.run_hook()
            self._run_started = time.time()

            th_fatalq_proc = Thread(target=self.thread_fatalq_processor)
            th_fatalq_proc.start()

            task_gens = []
            if (
                    self.master_taskq_cfg
                    and self.num_task_generators > 1
                ):
                raise IowebConfigError(
                    'num_task_generators must be 1 in'
                    ' master task queue mode'
                )
            for _ in range(self.num_task_generators):
                th = Thread(target=self.thread_task_generator)
                th.start()
                task_gens.append(th)

            th_debug = Thread(target=self.thread_debug)
            if self.debug:
                th_debug.start()

            pauses = [self.network_pause]
            result_workers = []
            for _ in range(self.result_workers):
                pause = Pause()
                th = Thread(
                    target=self.thread_result_processor,
                    args=[pause],
                ) 
                pauses.append(pause)
                th.start()
                result_workers.append(th)

            th_manager = Thread(
                target=self.thread_manager,
                args=[task_gens, pauses],
            )
            th_manager.start()

            th_network = Thread(
                target=self.thread_network,
            )
            th_network.start()

            th_manager.join()
            th_fatalq_proc.join()
            [x.join() for x in task_gens]
            if self.debug:
                th_debug.join()
            [x.join() for x in result_workers]

        except (Exception, KeyboardInterrupt):
            self.fatal_error_happened.set()
            raise
        finally:
            self.shutdown_event.set()
            self.shutdown()



"""
No gevent monkey patching is here.
Monkey patching is done by `ioweb` console command
which is imported from `ioweb_gevent` package
in setup.py entry point
"""
__version__ = '0.0.18'

from .cli import setup_logging
from .stat import Stat
from .session import Session, request
from .request import Request, CallbackRequest
from .data import Data
from .response import Response
from .crawler import Crawler
from .transport import Urllib3Transport
from .task_generator import TaskGenerator
from .error import *



from pprint import pprint
import sys
import re
import time
import os.path
#import json
from bson import json_util as json
import logging
from argparse import ArgumentParser
from importlib import import_module
from collections import defaultdict
from threading import Thread, Event
from subprocess import Popen, PIPE, STDOUT, TimeoutExpired
import code, traceback, signal

import psutil
from ioweb.stat import Stat

from pythonjsonlogger import jsonlogger

from .crawler import Crawler

logger = logging.getLogger('crawler.cli')

#from pympler import tracker
#mem_tracker = tracker.SummaryTracker()

def debug_handler(sig, frame):
    """Interrupt running process, and provide a python prompt for
    interactive debugging."""

    pass
    #import gc
    #from pympler import asizeof

    #objs = gc.get_objects()
    #mem = asizeof.asizeof(objs)
    #print('all objects memory: %s' % mem)

    #d={'_frame':frame, 'mem_tracker': mem_tracker}         # Allow access to frame object.
    #d.update(frame.f_globals)  # Unless shadowed by global
    #d.update(frame.f_locals)

    #i = code.InteractiveConsole(d)
    #message  = "Signal received : entering python shell.\nTraceback:\n"
    #message += ''.join(traceback.format_stack(frame))
    #i.interact(message)


def find_crawlers_in_module(mod, reg):
    for key in dir(mod):
        val = getattr(mod, key)
        if (
                isinstance(val, type)
                and issubclass(val, Crawler)
                and val is not Crawler
            ):
            logger.error(
                'Found crawler %s in module %s',
                val.__name__, mod.__file__
            )
            reg[val.__name__] = val


def collect_crawlers():
    reg = {}

    # Give crawlers in current directory max priority
    # Otherwise `/web/crawler/crawlers` packages are imported
    # when crawler is installed with `pip -e /web/crawler`
    sys.path.insert(0, os.getcwd())

    for location in ('crawlers',):
        try:
            mod = import_module(location)
        except ImportError as ex:
            #if path not in str(ex):
            logger.exception('Failed to import %s', location)
        else:
            if getattr(mod, '__file__', '').endswith('__init__.py'):
                dir_ = os.path.split(mod.__file__)[0]
                for fname in os.listdir(dir_):
                    if (
                        fname.endswith('.py')
                        and not fname.endswith('__init__.py')
                    ):
                        target_mod = '%s.%s' % (location, fname[:-3])
                        try:
                            mod = import_module(target_mod)
                        except ImportError as ex:
                            #if path not in str(ex):
                            logger.exception('Failed to import %s', target_mod)
                        else:
                            find_crawlers_in_module(mod, reg)
            else:
                find_crawlers_in_module(mod, reg)

    return reg


def setup_logging(logging_format='text', network_logs=False, verbose=False):#, control_logs=False):
    assert logging_format in ('text', 'json')

    root_logger = logging.getLogger()
    while root_logger.handlers:
        root_logger.handlers.pop()
    #logging.basicConfig(level=logging.DEBUG)
    if not verbose:
        logging.getLogger('urllib3.connectionpool').setLevel(level=logging.ERROR)
        logging.getLogger('urllib3.util.retry').setLevel(level=logging.ERROR)
        logging.getLogger('urllib3.poolmanager').setLevel(level=logging.ERROR)
        logging.getLogger('ioweb.urllib3_custom').setLevel(level=logging.ERROR)
        logging.getLogger('socks').setLevel(level=logging.INFO)
        if not network_logs:
            logging.getLogger('ioweb.network_service').propagate = False
    #if not control_logs:
    #    logging.getLogger('crawler.control').propagate = False

    hdl = logging.StreamHandler()
    logger = logging.getLogger()
    logger.addHandler(hdl)
    logger.setLevel(logging.DEBUG)
    if logging_format == 'json':
        #for hdl in logging.getLogger().handlers:
        hdl.setFormatter(jsonlogger.JsonFormatter())


def format_elapsed_time(total_sec):
    hours = minutes = 0
    if total_sec > 3600:
        hours, total_sec = divmod(total_sec, 3600)
    if total_sec > 60:
        minutes, total_sec = divmod(total_sec, 60)
    return '%02d:%02d:%.2f' % (hours, minutes, total_sec)


def get_crawler(crawler_id):
    reg = collect_crawlers()
    if crawler_id not in reg:
        sys.stderr.write(
            'Could not find %s crawler\n' % crawler_id
        )
        sys.exit(1)
    else:
        return reg[crawler_id]


def run_subcommand_crawl(opts):
    #logging.basicConfig(level=logging.DEBUG)
    setup_logging(
        logging_format=opts.logging_format,
        network_logs=opts.network_logs,
        verbose=opts.verbose
    )#, control_logs=opts.control_logs)
    cls = get_crawler(opts.crawler_id)
    extra_data = {}
    for key in cls.extra_cli_args():
        opt_key = 'extra_%s' % key.replace('-', '_')
        extra_data[key] = getattr(opts, opt_key)
    bot = cls(
        network_threads=opts.network_threads,
        extra_data=extra_data,
        debug=opts.debug,
        stat_logging=(opts.stat_logging == 'yes'),
        stat_logging_format=opts.stat_logging_format,
        master_taskq=(
            json.loads(opts.master_taskq)
            if opts.master_taskq else None
        ),
    )
    try:
        if opts.profile:
            import cProfile
            import pyprof2calltree
            import pstats

            profile_file = 'var/%s.prof' % opts.crawler_id
            profile_tree_file = 'var/%s.prof.out' % opts.crawler_id

            prof = cProfile.Profile()
            try:
                prof.runctx('bot.run()', globals(), locals())
            finally:
                stats = pstats.Stats(prof)
                stats.strip_dirs()
                pyprof2calltree.convert(stats, profile_tree_file)
        else:
            bot.run()
    except KeyboardInterrupt:
        logging.debug('Got INT signal')
        bot.fatal_error_happened.set()
    logging.debug('Stats:')
    for key, val in sorted(bot.stat.total_counters.items()):
        logging.debug(' * %s: %s' % (key, val))
    if bot._run_started:
        logging.debug('Elapsed: %s' % format_elapsed_time(time.time() - bot._run_started))
    else:
        logging.debug('Elapsed: NA')
    if bot.fatal_error_happened.is_set():
        sys.exit(1)
    else:
        sys.exit(0)


def get_master_taskq_id(crawler_id):
    return 'ioweb_taskq_%s' % crawler_id.lower()


def thread_worker(
        crawler_id, threads, stat, preg,
        evt_error, evt_init,
        master_taskq=False,
    ):
    counters = defaultdict(int)
    try:
        if master_taskq:
            taskq_id = get_master_taskq_id(crawler_id)
            master_taskq_cfg = json.dumps({
                'backend': 'redis',
                'taskq_id': taskq_id,
                'db': 13,
            })
        else:
            master_taskq_cfg = ''

        cmd = [
            'ioweb',
            'crawl',
            crawler_id,
            '-t%d' % threads,
            '--logging-format=json',
            '--stat-logging-format=json',
            '--master-taskq=%s' % master_taskq_cfg,
        ]
        proc = Popen(cmd, stdout=PIPE, stderr=STDOUT)
        preg[proc.pid] = proc
    finally:
        evt_init.set()
    while True:
        line = proc.stdout.readline()
        try:
            obj = json.loads(line.decode('ascii'))
        except (ValueError, UnicodeDecodeError):
            uline = line.decode('utf-8', errors='replace')
            logging.error(
                '[pid=%d] RAW-MSG: %s',
                proc.pid, uline.rstrip()
            )
        else:
            msg = obj['message']
            if 'exc_info' in obj:
                msg += obj['exc_info']
            try:
                msg_obj = json.loads(msg)
            except ValueError:
                logging.error('[pid=%d] TEXT-MSG: %s', proc.pid, msg)
            else:
                if 'eps' in msg_obj and 'counter' in msg_obj:
                    for key in msg_obj['eps'].keys():
                        stat.speed_keys = set(stat.speed_keys) | set([key])
                    for key, val in msg_obj['counter'].items():
                        delta = val - counters[key]
                        counters[key] = val
                        stat.inc(key, delta)
                else:
                    logging.error('[pid=%d] JSON-MSG: %s', proc.pid, msg_obj)
        ret = proc.poll()
        if ret is not None:
            if ret !=0:
                evt_error.set()
            break


def thread_task_gen(crawler_id, workers, evt_error):
    try:
        from redis import Redis
        rdb = Redis(db=13)
        taskq_id = get_master_taskq_id(crawler_id)
        cls = get_crawler(crawler_id)
        bot = cls(stat_logging=False)
        try:
            tgen_iter = iter(bot.task_generator())
        except TypeError:
            return
        else:
            try:
                taskq_cap = workers * bot.taskq_size_limit
                for task in tgen_iter:
                    while True:
                        if rdb.llen(taskq_id) < taskq_cap:
                            break
                        else:
                            time.sleep(0.01)
                    rdb.rpush(taskq_id, json.dumps(task.as_data()))
            finally:
                for _ in range(workers):
                    rdb.rpush(taskq_id, json.dumps({
                        '$op': 'shutdown',
                    }))
    except Exception:
        logging.exception('Error in thread_task_gen')
        evt_error.set()


def run_subcommand_multi(opts):
    setup_logging(
        logging_format='text',
        network_logs=True,
    )
    stat = Stat()

    pool = []
    preg = {}
    evt_error = Event()
    evt_stop = Event()
    try:
        if opts.master_taskq:
            th_task_gen = Thread(
                target=thread_task_gen,
                args=[opts.crawler_id, opts.workers, evt_error],
            )
            th_task_gen.daemon = True
            th_task_gen.start()

        for _ in range(opts.workers):
            evt_init = Event()
            th = Thread(
                target=thread_worker,
                args=[
                    opts.crawler_id,
                    opts.threads,
                    stat, preg, evt_error, evt_init,
                ],
                kwargs={
                    'master_taskq': opts.master_taskq,
                },
            )
            th.daemon = True
            th.start()
            pool.append(th)
            evt_init.wait()

        mem_check_time = time.time()

        while (
                not evt_stop.is_set()
                and not evt_error.is_set()
            ):
            num_done = 0
            for proc in preg.values():
                try:
                    proc.wait(timeout=0.1)
                except TimeoutExpired:
                    pass
                else:
                    num_done += 1
                if evt_error.is_set():
                    break
                if num_done == len(pool):
                    evt_stop.set()
                    break

                if opts.rss_limit:
                    if time.time() - mem_check_time > 5:
                        for proc in preg.values():
                            ps_proc = psutil.Process(proc.pid)
                            rss_mb = (
                                ps_proc.memory_info().rss
                                / (1024.0 * 1024)
                            )
                            if rss_mb > opts.rss_limit:
                                logging.error(
                                    'Terminating process, pid=%d, rss=%.02f mb, limit=%d mb' % (
                                        proc.pid, rss_mb, opts.rss_limit
                                    )
                                )
                                proc.terminate()
    finally:
        for proc in preg.values():
            logging.debug('[!] Finishing crawler subprocess, pid=%d' % proc.pid)
            proc.terminate()
            proc.wait()


def command_ioweb():
    #signal.signal(signal.SIGUSR1, debug_handler)

    parser = ArgumentParser()#add_help=False)

    crawler_cls = None
    if len(sys.argv) > 2:
        if sys.argv[1] == 'crawl':
            crawler_cls = get_crawler(sys.argv[2])

    subparsers = parser.add_subparsers(
        dest='command',
        title='Subcommands of ioweb command',
        description='',
    )

    # Crawl
    crawl_subparser = subparsers.add_parser(
        'crawl', help='Run crawler',
    )
    crawl_subparser.add_argument('crawler_id')
    crawl_subparser.add_argument('-t', '--network-threads', type=int, default=1)
    crawl_subparser.add_argument('-n', '--network-logs', action='store_true', default=False)
    crawl_subparser.add_argument('-p', '--profile', action='store_true', default=False)
    crawl_subparser.add_argument('--debug', action='store_true', default=False)
    crawl_subparser.add_argument('-v', '--verbose', action='store_true', default=False)
    crawl_subparser.add_argument(
        '--stat-logging', choices=['yes', 'no'], default='yes',
    )
    crawl_subparser.add_argument(
        '--stat-logging-format', choices=['text', 'json'], default='text',
    )
    crawl_subparser.add_argument(
        '--logging-format', choices=['text', 'json'], default='text',
    )
    #parser.add_argument('--control-logs', action='store_true', default=False)
    crawl_subparser.add_argument(
        '--master-taskq', type=str
    )
    if crawler_cls:
        crawler_cls.update_arg_parser(crawl_subparser)

    # multi
    multi_subparser = subparsers.add_parser(
        'multi', help='Run multi instances of crawler',
    )
    multi_subparser.add_argument('crawler_id')
    multi_subparser.add_argument(
        '-w', '--workers', type=int, default=1,
        help='Number of crawler processes to spawn',
    )
    multi_subparser.add_argument(
        '-t', '--threads', type=int, default=1,
        help='Number of thread in each crawler process',
    )
    multi_subparser.add_argument(
        '--master-taskq', action='store_true', default=False,
        help='Serve tasks from parent process',
    )
    multi_subparser.add_argument('--rss-limit', type=int)

    opts = parser.parse_args()
    if opts.command == 'crawl':
        run_subcommand_crawl(opts)
    elif opts.command == 'multi':
        run_subcommand_multi(opts)
    else:
        parser.print_help()
